@@ -38,6 +38,14 @@
 
   
    HISTORY: 
+   2006-09-30 : 2.4.2      - several code opimizations,
+                             PostScript PPD now uncompressed and rev4 only
+   2006-06-13 : 2.4.1      - added a PDF-to-mail postprocessing script to contrib/,
+                             added a new revision of the PostscriptColor PPD file
+   2006-05-28 : 2.4.0      - created the contrib/ directory with additional software
+   2006-05-17 : 2.3.1	   - fixed possible read access after end of string,
+                             fixed recognition of international characters 
+   2006-05-14 : 2.3.0	   - experimental option to decode hex strings in title (NK)
    2006-04-14 : 2.2.0      - new option to remove prefixes from usernames
    2006-04-06 : 2.1.1      - minor bugfix to read umasks properly
    2006-03-25 : 2.1.0	   - new option to select where to look for title,
@@ -133,14 +141,14 @@
 
 extern int errno;
 
-FILE *logfp=NULL;
+static FILE *logfp=NULL;
 
-void log_event(char type, char message[48], char *detail) {
+static void log_event(short type, char message[], char *detail) {
   time_t secs;
   char ctype[8], *timestring;
 
   if (strlen(conf.log) && (type & conf.logtype)) { 
-    time(&secs);
+    (void) time(&secs);
     timestring=ctime(&secs);
     timestring[strlen(timestring)-1]='\0';
 
@@ -163,7 +171,7 @@ void log_event(char type, char message[48], char *detail) {
   return;
 }
 
-void read_config() {
+static void read_config() {
   FILE *fp;
   cp_string buffer, key, value;
 
@@ -178,15 +186,15 @@ void read_config() {
       if (!strlen(key) || !strncmp(key,"#",1))
         continue;
       if (!_assign_value(key, value))
-        fputs("Error in configuration file!\n", stderr);
+        (void) fputs("Error in configuration file!\n", stderr);
     }
   }
 
-  fclose(fp);
+  (void) fclose(fp);
   return;
 }
 
-int create_dir(char *dirname, int nolog) {
+static int create_dir(char *dirname, int nolog) {
   struct stat fstatus;
   char buffer[BUFSIZE],*delim;
   int i;
@@ -194,7 +202,7 @@ int create_dir(char *dirname, int nolog) {
   while ((i=strlen(dirname))>1 && dirname[i-1]=='/')
     dirname[i-1]='\0';
   if (stat(dirname, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
-    memcpy(buffer,dirname,BUFSIZE-1);
+    strncpy(buffer,dirname,BUFSIZE);
     delim=strrchr(buffer,'/');
     if (delim!=buffer)
       delim[0]='\0';
@@ -202,7 +210,7 @@ int create_dir(char *dirname, int nolog) {
       delim[1]='\0';
     if (create_dir(buffer,nolog)!=0)
       return 1;
-    stat(buffer, &fstatus);
+    (void) stat(buffer, &fstatus);
     if (mkdir(dirname,fstatus.st_mode)!=0) {
       if (!nolog)
         log_event(CPERROR, "failed to create directory", dirname);
@@ -213,12 +221,12 @@ int create_dir(char *dirname, int nolog) {
         log_event(CPSTATUS, "directory created", dirname);
     if (chown(dirname,fstatus.st_uid,fstatus.st_gid)!=0)
       if (!nolog)
-        log_event(CPDEBUG, "failed to set mode on directory (non fatal)", dirname);
+        log_event(CPDEBUG, "failed to set owner on directory (non fatal)", dirname);
   }
   return 0;
 }
     
-int init() {
+static int init() {
   struct stat fstatus;
   struct group *group;
   cp_string logfile;
@@ -229,11 +237,11 @@ int init() {
         (S_ISREG(fstatus.st_mode) || S_ISLNK(fstatus.st_mode)))
       read_config();
 
-  umask(0077);
+  (void) umask(0077);
 
   group=getgrnam(conf.grp);
   if (group)
-    setgid(group->gr_gid);
+    (void) setgid(group->gr_gid);
 
   if (strlen(conf.log)) {
     if (stat(conf.log, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
@@ -254,14 +262,14 @@ int init() {
   else 
     log_event(CPDEBUG, "switching to new gid", conf.grp);
 
-  umask(0022);
+  (void) umask(0022);
 
   if (stat(conf.spool, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
     if (create_dir(conf.spool, 0)) {
       log_event(CPERROR, "failed to create spool directory", conf.spool);
       return 1;
     }
-    if (chmod(conf.spool, 0755)) {
+    if (chmod(conf.spool, 0751)) {
       log_event(CPERROR, "failed to set mode on spool directory", conf.spool);
       return 1;
     }
@@ -270,16 +278,16 @@ int init() {
     log_event(CPSTATUS, "spool directory created", conf.spool);
   }
 
-  umask(0077);
+  (void) umask(0077);
   return 0;
 }
 
-char *preparedirname(struct passwd *passwd, char *uname) {
+static char *preparedirname(struct passwd *passwd, char *uname) {
   int size;
   char bufin[BUFSIZE], bufout[BUFSIZE], *needle, *cptr;
 
   needle=strstr(uname, conf.removeprefix);
-  if (strlen(uname)>(unsigned int)(size=strlen(conf.removeprefix)))
+  if ((int)strlen(uname)>(size=strlen(conf.removeprefix)))
     uname=uname+size;
 
   strncpy(bufin, conf.out, BUFSIZE);
@@ -312,17 +320,17 @@ char *preparedirname(struct passwd *passwd, char *uname) {
   return cptr;
 }
 
-int prepareuser(struct passwd *passwd, char *dirname) {
+static int prepareuser(struct passwd *passwd, char *dirname) {
   struct stat fstatus;
 
-  umask(0000);
+  (void) umask(0000);
   if (stat(dirname, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
     if (!strcmp(passwd->pw_name, conf.anonuser)) {
       if (create_dir(dirname, 0)) {
         log_event(CPERROR, "failed to create anonymous output directory", dirname);
         return 1;
       }
-      if (chmod(dirname, (0777&~conf.anonumask))) {
+      if (chmod(dirname, (mode_t)(0777&~conf.anonumask))) {
         log_event(CPERROR, "failed to set mode on anonymous output directory", dirname);
         return 1;
       }
@@ -333,7 +341,7 @@ int prepareuser(struct passwd *passwd, char *dirname) {
         log_event(CPERROR, "failed to create user output directory", dirname);
         return 1;
       }
-      if (chmod(dirname, (0777&~conf.userumask))) {
+      if (chmod(dirname, (mode_t)(0777&~conf.userumask))) {
         log_event(CPERROR, "failed to set mode on user output directory", dirname);
         return 1;
       }
@@ -345,44 +353,142 @@ int prepareuser(struct passwd *passwd, char *dirname) {
     }
     log_event(CPDEBUG, "owner set for output directory", passwd->pw_name);
   }
-  umask(0077);
+  (void) umask(0077);
   return 0;
 }
 
-void replace_string(char *string) {
- unsigned int i;
+/* no validation is done here, please use is_ps_hex_string for that */
+static void decode_ps_hex_string(char *string) {
+  char *src_ptr, *dst_ptr;
+  int is_lower_digit; 					/* 0 - higher digit, 1 - lower digit */
+  char number, digit;
 
- log_event(CPDEBUG, "removing special characters from title", string);
- for (i=0;i<strlen(string);i++) 
-   if ( ( (int) string[i] < '0' || (int) string[i] > '9' ) &&
-        ( (int) string[i] < 'A' || (int) string[i] > 'Z' ) &&
-        ( (int) string[i] < 'a' || (int) string[i] > 'z' ) &&
-        (int) string[i] != '-' && (int) string[i] != '+' )
+  dst_ptr=string; 					/* we should always be behind src_ptr,
+                   			   		   so it's safe to write over original string */
+  number=(char)0;
+  is_lower_digit=0;
+  for (src_ptr=string+1;*src_ptr != '>';src_ptr++) { 	/* begin after start marker */
+    if (*src_ptr == ' ' || *src_ptr == '\t' ) {		/* skip whitespace */
+      continue;
+    }
+    if (*src_ptr >= 'a') {				/* assuming 0 < A < a */
+      digit=*src_ptr-'a'+(char)10;
+    }
+    else if (*src_ptr >= 'A') {
+      digit=*src_ptr-'A'+(char)10;
+    }
+    else {
+      digit=*src_ptr-'0';
+    }
+    if (is_lower_digit) {
+      number|=digit;
+      *dst_ptr=number;					/* write character */
+      dst_ptr++;
+      is_lower_digit=0;
+    }
+    else { 						/* higher digit */
+      number=digit<<4;
+      is_lower_digit=1;
+    }
+  }
+  if (is_lower_digit) {					/* write character with lower digit = 0,
+    							   as per PostScript Language Reference */
+    *dst_ptr=number;
+    dst_ptr++;
+    /* is_lower_digit=0; */
+  }
+  *dst_ptr=0;						/* finish him! */
+  return;
+}
+
+static int is_ps_hex_string(char *string) {
+  int got_end_marker=0;
+  char *ptr;
+  
+  if (string[0] != '<') { 	/* if has no start marker */
+    log_event(CPDEBUG, "not a hex string, has no start marker", string);
+    return 0;              		/* not hex string, obviously */
+  }
+  for (ptr=string+1;*ptr;ptr++) { 	/* begin after start marker */
+    if (got_end_marker) { 		/* got end marker and still something left */
+      log_event(CPDEBUG, "not a hex string, trailing characters after end marker", ptr);
+      return 0;           		/* that's bad! */
+    }
+    else if (*ptr == '>') { 	/* here it is! */
+      got_end_marker=1;
+      log_event(CPDEBUG, "got an end marker in the hex string, expecting 0-termination", ptr);
+    }
+    else if ( !(
+      isxdigit(*ptr) ||
+      *ptr == ' ' ||
+      *ptr == '\t'
+    ) ) {
+      log_event(CPDEBUG, "not a hex string, invalid character", ptr);
+      return 0;    			/* that's bad, too */
+    }
+  }
+  return got_end_marker;
+}
+
+static void alternate_replace_string(char *string) {
+  unsigned int i;
+
+  log_event(CPDEBUG, "removing alternate special characters from title", string);
+  for (i=0;i<(unsigned int)strlen(string);i++)
+   if ( isascii(string[i]) &&      	/* leaving non-ascii characters intact */
+        (!isalnum(string[i])) && 
+        string[i] != '-' && string[i] != '+' )
     string[i]='_';
   return;
 }
 
-int preparetitle(char *title) {
+static void replace_string(char *string) {
+  unsigned int i;
+
+  log_event(CPDEBUG, "removing special characters from title", string);
+  for (i=0;i<(unsigned int)strlen(string);i++) 
+    if ( ( string[i] < '0' || string[i] > '9' ) &&
+         ( string[i] < 'A' || string[i] > 'Z' ) &&
+         ( string[i] < 'a' || string[i] > 'z' ) &&
+         string[i] != '-' && string[i] != '+' )
+      string[i]='_';
+  return;
+}
+
+static int preparetitle(char *title) {
   char *cut;
   
+  if (title != NULL) {
+    log_event(CPDEBUG, "removing trailing newlines from title", title);
+    while (strlen(title) && ((title[strlen(title)-1] == '\n') || (title[strlen(title)-1] == '\r')))
+      title[strlen(title)-1]='\0';
+  }
+  if (conf.decodehexstrings) {
+    log_event(CPSTATUS, "***Experimental Option: DecodeHexStrings", NULL);
+    log_event(CPDEBUG, "checking for hex strings", title);
+    if (is_ps_hex_string(title))
+      decode_ps_hex_string(title);
+  }
   cut=strrchr(title, '/');
   if (cut != NULL) {
     log_event(CPDEBUG, "removing slashes from full title", title);
     memmove(title, cut+1, strlen(cut+1)+1);
   }  
-  if (title != NULL) 
-    while (strlen(title) && ((title[strlen(title)-1] == '\n') || (title[strlen(title)-1] == '\r'))) 
-      title[strlen(title)-1]='\0';
   cut=strrchr(title, '.');
-  if ((cut != NULL) && ((short)strlen(cut) <= conf.cut+1) && (cut != title)) {
+  if ((cut != NULL) && ((int)strlen(cut) <= conf.cut+1) && (cut != title)) {
     log_event(CPDEBUG, "removing file name extension", cut);
     cut[0]='\0';
   }
-  replace_string(title);
+  if (conf.decodehexstrings) {
+    log_event(CPDEBUG, "calling alternate_replace_string", NULL);
+    alternate_replace_string(title);
+  }
+  else
+    replace_string(title);
   return strcmp(title, "");
 }
 
-int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle, 
+static int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle, 
                      int job, struct passwd *passwd) {
   cp_string buffer;
   int rec_depth,is_title=0;
@@ -396,7 +502,7 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle,
   fpdest=fopen(spoolfile, "w");
   if (fpdest == NULL) {
     log_event(CPERROR, "failed to open spoolfile", spoolfile);
-    fclose(fpsrc);
+    (void) fclose(fpsrc);
     return 1;
   }
   log_event(CPDEBUG, "destination stream ready", spoolfile);
@@ -413,9 +519,9 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle,
     }
   }
   log_event(CPDEBUG, "now extracting postscript code", NULL);
-  fputs(buffer, fpdest);
+  (void) fputs(buffer, fpdest);
   while (fgets(buffer, BUFSIZE, fpsrc) != NULL) {
-    fputs(buffer, fpdest);
+    (void) fputs(buffer, fpdest);
     if (!is_title && !rec_depth)
       if (sscanf(buffer, "%%%%Title: %"TBUFSIZE"c", title)==1) {
         log_event(CPDEBUG, "found title in ps code", title);
@@ -436,8 +542,8 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle,
       }
     }
   }
-  fclose(fpdest);
-  fclose(fpsrc);
+  (void) fclose(fpdest);
+  (void) fclose(fpsrc);
   log_event(CPDEBUG, "all data written to spoolfile", spoolfile);
 
   if (!strcmp(cmdtitle, "(stdin)")) 
@@ -486,7 +592,7 @@ int main(int argc, char *argv[]) {
   pid_t pid;
 
   if (setuid(0)) {
-    fputs("cups-pdf cannot be called without root privileges!\n", stderr);
+    (void) fputs("cups-pdf cannot be called without root privileges!\n", stderr);
     return 0;
   }
 
@@ -500,7 +606,7 @@ int main(int argc, char *argv[]) {
     return 0;
   }
   if (argc<6 || argc>7) {
-    fputs("Usage: cups-pdf job-id user title copies options [file]\n", stderr);
+    (void) fputs("Usage: cups-pdf job-id user title copies options [file]\n", stderr);
     log_event(CPERROR, "call contained illegal number of arguments", NULL);
     return 0;
   }
@@ -508,7 +614,7 @@ int main(int argc, char *argv[]) {
   size=strlen(conf.userprefix)+strlen(argv[2])+1;
   user=calloc(size, sizeof(char));
   if (user == NULL) {
-    fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+    (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     return -2;
   }  
   snprintf(user, size, "%s%s", conf.userprefix, argv[2]);
@@ -529,17 +635,17 @@ int main(int argc, char *argv[]) {
         log_event(CPERROR, "username for anonymous access unknown", conf.anonuser);
         free(user);
         if (logfp!=NULL)
-          fclose(logfp);
+          (void) fclose(logfp);
         return -2;
       }
       log_event(CPDEBUG, "unknown user", user);
       size=strlen(conf.anondirname)+4;
       dirname=calloc(size, sizeof(char));
       if (dirname == NULL) {
-        fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+        (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
         free(user);
         if (logfp!=NULL)
-          fclose(logfp);
+          (void) fclose(logfp);
         return -2;
       }  
       snprintf(dirname, size, "%s", conf.anondirname);
@@ -552,31 +658,31 @@ int main(int argc, char *argv[]) {
       log_event(CPSTATUS, "anonymous access denied", user);
       free(user);
       if (logfp!=NULL)
-        fclose(logfp);
+        (void) fclose(logfp);
       return 0;
     }
-    mode=(0666&~conf.anonumask);
+    mode=(mode_t)(0666&~conf.anonumask);
   } 
   else {
     log_event(CPDEBUG, "user identified", passwd->pw_name);
     if ((dirname=preparedirname(passwd, argv[2])) == NULL) {
-      fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+      (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
       free(user);
       if (logfp!=NULL)
-        fclose(logfp);
+        (void) fclose(logfp);
       return -2;
     }  
     while (strlen(dirname) && ((dirname[strlen(dirname)-1] == '\n') ||
            (dirname[strlen(dirname)-1] == '\r')))
       dirname[strlen(dirname)-1]='\0';
     log_event(CPDEBUG, "output directory name generated", dirname);
-    mode=(0666&~conf.userumask);
+    mode=(mode_t)(0666&~conf.userumask);
   }
   free(user);
   if (prepareuser(passwd, dirname)) {
     free(dirname);
     if (logfp!=NULL)
-      fclose(logfp);
+      (void) fclose(logfp);
     return -2;
   }
   log_event(CPDEBUG, "user information prepared", NULL);
@@ -584,10 +690,10 @@ int main(int argc, char *argv[]) {
   size=strlen(conf.spool)+22;
   spoolfile=calloc(size, sizeof(char));
   if (spoolfile == NULL) {
-    fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+    (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     free(dirname);
     if (logfp!=NULL)
-      fclose(logfp);
+      (void) fclose(logfp);
     return -3;
   }
   snprintf(spoolfile, size, "%s/cups2pdf-%i", conf.spool, (int) getpid());
@@ -598,7 +704,7 @@ int main(int argc, char *argv[]) {
       free(dirname);
       free(spoolfile);
       if (logfp!=NULL)
-        fclose(logfp);
+        (void) fclose(logfp);
       return -4;
     }
     log_event(CPDEBUG, "input data read from stdin", NULL);
@@ -608,7 +714,7 @@ int main(int argc, char *argv[]) {
       free(dirname);
       free(spoolfile);
       if (logfp!=NULL)
-        fclose(logfp);
+        (void) fclose(logfp);
       return -4;
     }
     log_event(CPDEBUG, "input data read from file", argv[6]);
@@ -617,11 +723,11 @@ int main(int argc, char *argv[]) {
   size=strlen(dirname)+strlen(title)+6;
   outfile=calloc(size, sizeof(char));
   if (outfile == NULL) {
-    fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+    (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     free(dirname);
     free(spoolfile);
     if (logfp!=NULL)
-      fclose(logfp);
+      (void) fclose(logfp);
     return -5;
   }
   snprintf(outfile, size, "%s/%s.pdf", dirname, title);
@@ -630,18 +736,18 @@ int main(int argc, char *argv[]) {
   size=strlen(conf.gscall)+strlen(conf.ghostscript)+strlen(conf.pdfver)+strlen(outfile)+strlen(spoolfile)+6;
   gscall=calloc(size, sizeof(char));
   if (gscall == NULL) {
-    fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+    (void) fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     free(dirname);
     free(spoolfile);
     free(outfile);
     if (logfp!=NULL)
-      fclose(logfp);
+      (void) fclose(logfp);
     return -6;
   }
   snprintf(gscall, size, conf.gscall, conf.ghostscript, conf.pdfver, outfile, spoolfile);
   log_event(CPDEBUG, "ghostscript commandline built", gscall);
 
-  unlink(outfile);
+  (void) unlink(outfile);
   log_event(CPDEBUG, "output file unlinked", outfile);
 
   if (setenv("TMPDIR", conf.gstmp, 1)) {
@@ -651,13 +757,13 @@ int main(int argc, char *argv[]) {
     free(outfile);
     free(gscall);
     if (logfp!=NULL)
-      fclose(logfp);
+      (void) fclose(logfp);
     return -7;
   }
   log_event(CPDEBUG, "TMPDIR set for GhostScript", getenv("TMPDIR"));
 
   if (logfp!=NULL)
-    fflush(logfp);
+    (void) fflush(logfp);
   pid=fork();
 
   if (!pid) {
@@ -672,7 +778,7 @@ int main(int argc, char *argv[]) {
     else 
       log_event(CPDEBUG, "UID set for current user", passwd->pw_name);
      
-    umask(0077);
+    (void) umask(0077);
     size=system(gscall);
     snprintf(title,BUFSIZE,"%d",size);
     log_event(CPDEBUG, "ghostscript has finished", title);
@@ -702,7 +808,7 @@ int main(int argc, char *argv[]) {
     return 0;
   }
   log_event(CPDEBUG, "waiting for child to exit", NULL);
-  waitpid(pid,NULL,0);
+  (void) waitpid(pid,NULL,0);
 
   if (unlink(spoolfile)) 
     log_event(CPERROR, "failed to unlink spoolfile (non fatal)", spoolfile);
@@ -719,6 +825,6 @@ int main(int argc, char *argv[]) {
   log_event(CPSTATUS, "PDF creation successfully finished", passwd->pw_name);
 
   if (logfp!=NULL)
-    fclose(logfp);
+    (void) fclose(logfp);
   return 0;
 } 
